import asyncio
from typing import List, Optional
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.postprocessor.types import BaseNodePostprocessor
from llama_index.core.async_utils import run_async_tasks
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.schema import NodeWithScore
from llama_index.core.indices.query.schema import QueryBundle
from llama_index.networks.contributor.retriever import ContributorRetrieverClient
from functools import reduce


class NetworkRetriever(BaseRetriever):
    """The network Retriever."""

    def __init__(
        self,
        contributors: List[ContributorRetrieverClient],
        node_postprocessors: Optional[List[BaseNodePostprocessor]] = None,
        callback_manager: Optional[CallbackManager] = None,
    ) -> None:
        self._contributors = contributors
        self._node_postprocessors = node_postprocessors or []
        super().__init__(callback_manager=callback_manager)

    def _postprocess_nodes(
        self, nodes: List[NodeWithScore], query_bundle: QueryBundle
    ) -> List[NodeWithScore]:
        return self.reranker.postprocess_nodes(nodes=nodes, query_bundle=query_bundle)

    def _retrieve(self, query_bundle: QueryBundle) -> List[NodeWithScore]:
        async_tasks = [
            contributor.aretrieve(query_bundle) for contributor in self._contributors
        ]
        contributor_results: List[List[NodeWithScore]] = run_async_tasks(async_tasks)
        nodes = reduce(lambda x, y: x + y, contributor_results)

        for postprocessor in self._node_postprocessors:
            nodes = postprocessor.postprocess_nodes(nodes, query_bundle=query_bundle)

        return nodes

    async def _aretrieve(self, query_bundle: QueryBundle) -> List[NodeWithScore]:
        async_tasks = [
            contributor.aretrieve(query_bundle) for contributor in self._contributors
        ]
        contributor_results: List[List[NodeWithScore]] = await asyncio.gather(
            *async_tasks
        )
        nodes = reduce(lambda x, y: x + y, contributor_results)

        for postprocessor in self._node_postprocessors:
            nodes = postprocessor.postprocess_nodes(nodes, query_bundle=query_bundle)

        return nodes

    @property
    def contributors(self) -> List[ContributorRetrieverClient]:
        """Get list of contributors."""
        return self._contributors



from llama_index.networks.contributor.retriever.service import (
    ContributorRetrieverService,
    ContributorRetrieverServiceSettings,
)
from llama_index.networks.contributor.retriever.client import (
    ContributorRetrieverClient,
    ContributorRetrieverClientSettings,
)


__all__ = [
    "ContributorRetrieverService",
    "ContributorRetrieverServiceSettings",
    "ContributorRetrieverClient",
    "ContributorRetrieverClientSettings",
]



from llama_index.networks.network.query_engine import NetworkQueryEngine
from llama_index.networks.network.retriever import NetworkRetriever

__all__ = [
    "ContributorClient",
    "ContributorClientSettings",
    "ContributorService",
    "ContributorServiceSettings",
    "NetworkQueryEngine",
    "NetworkRetriever",
]



from llama_index.networks.contributor.query_engine.service import (
    ContributorQueryEngineService,
    ContributorQueryEngineServiceSettings,
    ContributorService,
    ContributorServiceSettings,
)
from llama_index.networks.contributor.query_engine.client import (
    ContributorQueryEngineClient,
    ContributorQueryEngineClientSettings,
    ContributorClientSettings,
    ContributorClient,
)


__all__ = [
    "ContributorQueryEngineService",
    "ContributorQueryEngineServiceSettings",
    "ContributorQueryEngineClient",
    "ContributorQueryEngineClientSettings",
    "ContributorService",
    "ContributorServiceSettings",
    "ContributorClient",
    "ContributorClientSettings",
]



from typing import List, Optional
from llama_index.core.async_utils import run_async_tasks
from llama_index.core.bridge.pydantic import BaseModel
from llama_index.core.base.base_query_engine import BaseQueryEngine
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.callbacks.schema import CBEventType, EventPayload
from llama_index.core.base.response.schema import RESPONSE_TYPE
from llama_index.core.schema import TextNode, NodeWithScore, QueryBundle
from llama_index.core.response_synthesizers import (
    BaseSynthesizer,
    ResponseMode,
    get_response_synthesizer,
)
from llama_index.core.settings import (
    Settings,
)
from llama_index.core.prompts import BasePromptTemplate
from llama_index.core.prompts.mixin import PromptMixinType
from llama_index.core.llms import LLM
from llama_index.networks.contributor.query_engine import ContributorQueryEngineClient
import asyncio


class NetworkQueryEngine(BaseQueryEngine):
    """The network Query Engine."""

    def __init__(
        self,
        contributors: List[ContributorQueryEngineClient],
        response_synthesizer: Optional[BaseSynthesizer] = None,
        callback_manager: Optional[CallbackManager] = None,
    ) -> None:
        self._contributors = contributors
        self._response_synthesizer = response_synthesizer or get_response_synthesizer(
            llm=Settings.llm, callback_manager=Settings.callback_manager
        )
        super().__init__(callback_manager=callback_manager)

    @classmethod
    def from_args(
        cls,
        contributors: List[ContributorQueryEngineClient],
        llm: Optional[LLM] = None,
        response_synthesizer: Optional[BaseSynthesizer] = None,
        # response synthesizer args
        response_mode: ResponseMode = ResponseMode.COMPACT,
        text_qa_template: Optional[BasePromptTemplate] = None,
        refine_template: Optional[BasePromptTemplate] = None,
        summary_template: Optional[BasePromptTemplate] = None,
        simple_template: Optional[BasePromptTemplate] = None,
        output_cls: Optional[BaseModel] = None,
        use_async: bool = False,
        streaming: bool = False,
    ) -> "NetworkQueryEngine":
        llm = llm or Settings.llm

        response_synthesizer = response_synthesizer or get_response_synthesizer(
            llm=llm,
            text_qa_template=text_qa_template,
            refine_template=refine_template,
            summary_template=summary_template,
            simple_template=simple_template,
            response_mode=response_mode,
            output_cls=output_cls,
            use_async=use_async,
            streaming=streaming,
        )

        callback_manager = Settings.callback_manager

        return cls(
            contributors=contributors,
            response_synthesizer=response_synthesizer,
            callback_manager=callback_manager,
        )

    def _get_prompt_modules(self) -> PromptMixinType:
        """Get prompt sub-modules."""
        return {"response_synthesizer": self._response_synthesizer}

    def _query(self, query_bundle: QueryBundle) -> RESPONSE_TYPE:
        """Answer a query."""
        with self.callback_manager.event(
            CBEventType.QUERY, payload={EventPayload.QUERY_STR: query_bundle.query_str}
        ) as query_event:
            results = []
            async_tasks = [
                contributor.aquery(query_bundle) for contributor in self._contributors
            ]
            results = run_async_tasks(async_tasks)

            nodes = [
                NodeWithScore(
                    node=TextNode(text=el.response), score=el.metadata["score"]
                )
                for el in results
            ]
            response = self._response_synthesizer.synthesize(
                query=query_bundle,
                nodes=nodes,
            )

            query_event.on_end(payload={EventPayload.RESPONSE: response})

        return response

    async def _aquery(self, query_bundle: QueryBundle) -> RESPONSE_TYPE:
        """Answer a query."""
        with self.callback_manager.event(
            CBEventType.QUERY, payload={EventPayload.QUERY_STR: query_bundle.query_str}
        ) as query_event:
            # go through all clients
            query_tasks = []
            for contributor in self._contributors:
                query_tasks += [contributor.aquery(query_bundle)]

            results = await asyncio.gather(*query_tasks)
            nodes = [
                NodeWithScore(
                    node=TextNode(text=el.response), score=el.metadata["score"]
                )
                for el in results
            ]
            response = await self._response_synthesizer.asynthesize(
                query=query_bundle,
                nodes=nodes,
            )

            query_event.on_end(payload={EventPayload.RESPONSE: response})

        return response

    @property
    def contributors(self) -> List[ContributorQueryEngineClient]:
        """Get the retriever object."""
        return self._contributors



from typing import Optional, List, Dict, Any, Type
from llama_index.core.schema import (
    NodeWithScore,
    TextNode,
    BaseNode,
    IndexNode,
    ImageNode,
)
from llama_index.core.base.response.schema import Response
from llama_index.core.bridge.pydantic import BaseModel
from pydantic import BaseModel as V2BaseModel

NODE_REGISTRY: Dict[str, Type[BaseNode]] = {
    "TextNode": TextNode,
    "IndexNode": IndexNode,
    "ImageNode": ImageNode,
}


class ContributorQueryRequest(V2BaseModel):
    query: str


class ContributorQueryResponse(BaseModel):
    response: Optional[str]
    score: Optional[float]

    def __str__(self) -> str:
        """Convert to string representation."""
        return self.response or "None"

    def to_response(self) -> Response:
        """Convert to Response type."""
        return Response(response=self.response, metadata={"score": self.score})


class ContributorRetrieverRequest(V2BaseModel):
    query: str


class ContributorRetrieverResponse(BaseModel):
    nodes_dict: Optional[List[Dict[str, Any]]]

    def get_nodes(self) -> List[NodeWithScore]:
        """Build list of nodes with score."""
        nodes = []
        for d in self.nodes_dict:
            node_dict = d["node"]
            try:
                node_cls = NODE_REGISTRY[node_dict["class_name"]]
            except KeyError:
                node_cls = NODE_REGISTRY["TextNode"]
            node = node_cls.parse_obj(node_dict)
            nodes.append(NodeWithScore(node=node, score=d["score"]))
        return nodes



from typing import Optional, Dict, List
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.schema import QueryBundle, NodeWithScore
from llama_index.core.prompts.mixin import PromptMixinType
from llama_index.networks.schema.contributor import ContributorRetrieverResponse
from pydantic.v1 import BaseSettings, Field
import requests
import aiohttp


class ContributorRetrieverClientSettings(BaseSettings):
    """Settings for contributor."""

    api_key: Optional[str] = Field(default=None, env="API_KEY")
    api_url: str = Field(..., env="API_URL")

    class Config:
        env_file = ".env", ".env.contributor.client"


class ContributorRetrieverClient(BaseRetriever):
    """A remote Retriever exposed through a REST API."""

    def __init__(
        self,
        callback_manager: Optional[CallbackManager],
        config: ContributorRetrieverClientSettings,
    ) -> None:
        self.config = config
        super().__init__(callback_manager)

    @classmethod
    def from_config_file(
        cls, env_file: str, callback_manager: Optional[CallbackManager] = None
    ) -> "ContributorRetrieverClient":
        """Convenience constructor from a custom env file."""
        config = ContributorRetrieverClientSettings(_env_file=env_file)
        return cls(callback_manager=callback_manager, config=config)

    def _retrieve(
        self,
        query_bundle: QueryBundle,
        additional_data: Dict[str, str] = {},
        headers: Dict[str, str] = {},
    ) -> List[NodeWithScore]:
        """Make a post request to submit a query to Retriever."""
        # headers = {"Authorization": f"Bearer {self.config.api_key}"}
        data = {"query": query_bundle.query_str, "api_key": self.config.api_key}
        data.update(additional_data)
        result = requests.post(
            self.config.api_url + "/api/retrieve", json=data, headers=headers
        )
        try:
            contributor_response = ContributorRetrieverResponse.parse_obj(result.json())
        except Exception as e:
            raise ValueError("Failed to parse response") from e
        return contributor_response.get_nodes()

    async def _aretrieve(
        self,
        query_bundle: QueryBundle,
        api_token: Optional[str] = None,
        additional_data: Dict[str, str] = {},
        headers: Dict[str, str] = {},
    ) -> List[NodeWithScore]:
        """Make a post request to submit a query to Retriever."""
        # headers = {"Authorization": f"Bearer {self.config.api_key}"}
        data = {"query": query_bundle.query_str, "api_token": api_token}
        data.update(additional_data)
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.config.api_url + "/api/retrieve", json=data, headers=headers
            ) as resp:
                json_result = await resp.json()
            try:
                contributor_response = ContributorRetrieverResponse.parse_obj(
                    json_result
                )
            except Exception as e:
                raise ValueError("Failed to parse response") from e
        return contributor_response.get_nodes()

    def _get_prompt_modules(self) -> PromptMixinType:
        """Get prompt sub-modules."""
        return {}



from llama_index.networks.contributor.query_engine import (
    ContributorClient,
    ContributorClientSettings,
    ContributorService,
    ContributorServiceSettings,
)


__all__ = [
    "ContributorClient",
    "ContributorClientSettings",
    "ContributorService",
    "ContributorServiceSettings",
]



from typing import Any, Optional
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.bridge.pydantic import Field, BaseModel
from llama_index.networks.schema.contributor import (
    ContributorRetrieverRequest,
)
from pydantic.v1 import BaseSettings, PrivateAttr
from fastapi import FastAPI


class ContributorRetrieverServiceSettings(BaseSettings):
    api_version: str = Field(default="v1", description="API version.")
    secret: Optional[str] = Field(
        default=None, description="JWT secret."
    )  # left for future consideration.
    # or if user wants to implement their own

    class Config:
        env_file = ".env", ".env.contributor.service"


class ContributorRetrieverService(BaseModel):
    retriever: Optional[BaseRetriever]
    config: ContributorRetrieverServiceSettings
    _fastapi: FastAPI = PrivateAttr()

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, retriever, config) -> None:
        self._fastapi = FastAPI(
            version=config.api_version,
        )

        # routes
        self._fastapi.add_api_route(path="/api", endpoint=self.index, methods=["GET"])
        self._fastapi.add_api_route(
            path="/api/retrieve",
            endpoint=self.retrieve,
            methods=["POST"],
        )

        super().__init__(retriever=retriever, config=config)

    async def index(self):
        """Index endpoint logic."""
        return {"message": "Hello World!"}

    async def retrieve(self, request: ContributorRetrieverRequest):
        """Retriever endpoint logic."""
        result = await self.retriever.aretrieve(request.query)
        return {
            "nodes_dict": result,
        }

    @classmethod
    def from_config_file(
        cls, env_file: str, retriever: BaseRetriever
    ) -> "ContributorRetrieverService":
        config = ContributorRetrieverServiceSettings(_env_file=env_file)
        return cls(retriever=retriever, config=config)

    def __getattr__(self, attr) -> Any:
        if hasattr(self._fastapi, attr):
            return getattr(self._fastapi, attr)
        else:
            raise AttributeError(f"{attr} not exist")

    @property
    def app(self):
        return self._fastapi
